@@ -20,36 +20,47 @@ information or mock the environment.
 
 - `has_program` checks the presence of a command on the system.
 
-- The `generate_album_info` and `generate_track_info` functions return
-  fixtures to be used when mocking the autotagger.
-
 - The `ImportSessionFixture` allows one to run importer code while
   controlling the interactions through code.
 
 - The `TestHelper` class encapsulates various fixtures that can be set up.
 """
 
+from __future__ import annotations
 
 import os
 import os.path
 import shutil
 import subprocess
 import sys
+import unittest
 from contextlib import contextmanager
+from dataclasses import dataclass
 from enum import Enum
+from functools import cached_property
 from io import StringIO
-from tempfile import mkdtemp, mkstemp
+from pathlib import Path
+from tempfile import gettempdir, mkdtemp, mkstemp
+from typing import Any, ClassVar
+from unittest.mock import patch
 
+import responses
 from mediafile import Image, MediaFile
 
 import beets
 import beets.plugins
-from beets import autotag, config, importer, logging, util
+from beets import importer, logging, util
 from beets.autotag.hooks import AlbumInfo, TrackInfo
-from beets.library import Album, Item, Library
+from beets.importer import ImportSession
+from beets.library import Item, Library
 from beets.test import _common
 from beets.ui.commands import TerminalImportSession
-from beets.util import MoveOperation, bytestring_path, syspath
+from beets.util import (
+    MoveOperation,
+    bytestring_path,
+    clean_module_tempdir,
+    syspath,
+)
 
 
 class LogCapture(logging.Handler):
@@ -107,20 +118,9 @@ def capture_stdout():
         print(capture.getvalue())
 
 
-def _convert_args(args):
-    """Convert args to bytestrings for Python 2 and convert them to strings
-    on Python 3.
-    """
-    for i, elem in enumerate(args):
-        if isinstance(elem, bytes):
-            args[i] = elem.decode(util.arg_encoding())
-
-    return args
-
-
 def has_program(cmd, args=["--version"]):
     """Returns `True` if `cmd` can be executed."""
-    full_cmd = _convert_args([cmd] + args)
+    full_cmd = [cmd] + args
     try:
         with open(os.devnull, "wb") as devnull:
             subprocess.check_call(
@@ -134,16 +134,81 @@ def has_program(cmd, args=["--version"]):
         return True
 
 
-class TestHelper:
+def check_reflink_support(path: str) -> bool:
+    try:
+        import reflink
+    except ImportError:
+        return False
+
+    return reflink.supported_at(path)
+
+
+class ConfigMixin:
+    @cached_property
+    def config(self) -> beets.IncludeLazyConfig:
+        """Base beets configuration for tests."""
+        config = beets.config
+        config.sources = []
+        config.read(user=False, defaults=True)
+
+        config["plugins"] = []
+        config["verbose"] = 1
+        config["ui"]["color"] = False
+        config["threaded"] = False
+        return config
+
+
+NEEDS_REFLINK = unittest.skipUnless(
+    check_reflink_support(gettempdir()), "no reflink support for libdir"
+)
+
+
+class IOMixin:
+    @cached_property
+    def io(self) -> _common.DummyIO:
+        return _common.DummyIO()
+
+    def setUp(self):
+        super().setUp()
+        self.io.install()
+
+    def tearDown(self):
+        super().tearDown()
+        self.io.restore()
+
+
+class TestHelper(ConfigMixin):
     """Helper mixin for high-level cli and plugin tests.
 
     This mixin provides methods to isolate beets' global state provide
     fixtures.
     """
 
+    resource_path = Path(os.fsdecode(_common.RSRC)) / "full.mp3"
+
+    db_on_disk: ClassVar[bool] = False
+
+    @cached_property
+    def temp_dir_path(self) -> Path:
+        return Path(self.create_temp_dir())
+
+    @cached_property
+    def temp_dir(self) -> bytes:
+        return util.bytestring_path(self.temp_dir_path)
+
+    @cached_property
+    def lib_path(self) -> Path:
+        lib_path = self.temp_dir_path / "libdir"
+        lib_path.mkdir(exist_ok=True)
+        return lib_path
+
+    @cached_property
+    def libdir(self) -> bytes:
+        return bytestring_path(self.lib_path)
+
     # TODO automate teardown through hook registration
 
-    def setup_beets(self, disk=False):
+    def setup_beets(self):
         """Setup pristine global configuration and library for testing.
 
         Sets ``beets.config`` so we can safely use any functionality
@@ -158,129 +223,33 @@ class TestHelper:
         - ``libdir`` Path to a subfolder of ``temp_dir``, containing the
           library's media files. Same as ``config['directory']``.
 
-        - ``config`` The global configuration used by beets.
-
         - ``lib`` Library instance created with the settings from
           ``config``.
 
         Make sure you call ``teardown_beets()`` afterwards.
         """
-        self.create_temp_dir()
-        os.environ["BEETSDIR"] = util.py3_path(self.temp_dir)
+        temp_dir_str = str(self.temp_dir_path)
+        self.env_patcher = patch.dict(
+            "os.environ",
+            {
+                "BEETSDIR": temp_dir_str,
+                "HOME": temp_dir_str,  # used by Confuse to create directories.
+            },
+        )
+        self.env_patcher.start()
 
-        self.config = beets.config
-        self.config.clear()
-        self.config.read()
+        self.config["directory"] = str(self.lib_path)
 
-        self.config["plugins"] = []
-        self.config["verbose"] = 1
-        self.config["ui"]["color"] = False
-        self.config["threaded"] = False
-
-        self.libdir = os.path.join(self.temp_dir, b"libdir")
-        os.mkdir(syspath(self.libdir))
-        self.config["directory"] = util.py3_path(self.libdir)
-
-        if disk:
+        if self.db_on_disk:
             dbpath = util.bytestring_path(self.config["library"].as_filename())
         else:
             dbpath = ":memory:"
         self.lib = Library(dbpath, self.libdir)
 
     def teardown_beets(self):
+        self.env_patcher.stop()
         self.lib._close()
-        if "BEETSDIR" in os.environ:
-            del os.environ["BEETSDIR"]
         self.remove_temp_dir()
-        self.config.clear()
-        beets.config.read(user=False, defaults=True)
-
-    def load_plugins(self, *plugins):
-        """Load and initialize plugins by names.
-
-        Similar setting a list of plugins in the configuration. Make
-        sure you call ``unload_plugins()`` afterwards.
-        """
-        # FIXME this should eventually be handled by a plugin manager
-        beets.config["plugins"] = plugins
-        beets.plugins.load_plugins(plugins)
-        beets.plugins.find_plugins()
-
-        # Take a backup of the original _types and _queries to restore
-        # when unloading.
-        Item._original_types = dict(Item._types)
-        Album._original_types = dict(Album._types)
-        Item._types.update(beets.plugins.types(Item))
-        Album._types.update(beets.plugins.types(Album))
-
-        Item._original_queries = dict(Item._queries)
-        Album._original_queries = dict(Album._queries)
-        Item._queries.update(beets.plugins.named_queries(Item))
-        Album._queries.update(beets.plugins.named_queries(Album))
-
-    def unload_plugins(self):
-        """Unload all plugins and remove the from the configuration."""
-        # FIXME this should eventually be handled by a plugin manager
-        beets.config["plugins"] = []
-        beets.plugins._classes = set()
-        beets.plugins._instances = {}
-        Item._types = Item._original_types
-        Album._types = Album._original_types
-        Item._queries = Item._original_queries
-        Album._queries = Album._original_queries
-
-    def create_importer(self, item_count=1, album_count=1):
-        """Create files to import and return corresponding session.
-
-        Copies the specified number of files to a subdirectory of
-        `self.temp_dir` and creates a `ImportSessionFixture` for this path.
-        """
-        import_dir = os.path.join(self.temp_dir, b"import")
-        if not os.path.isdir(syspath(import_dir)):
-            os.mkdir(syspath(import_dir))
-
-        album_no = 0
-        while album_count:
-            album = util.bytestring_path(f"album {album_no}")
-            album_dir = os.path.join(import_dir, album)
-            if os.path.exists(syspath(album_dir)):
-                album_no += 1
-                continue
-            os.mkdir(syspath(album_dir))
-            album_count -= 1
-
-            track_no = 0
-            album_item_count = item_count
-            while album_item_count:
-                title = f"track {track_no}"
-                src = os.path.join(_common.RSRC, b"full.mp3")
-                title_file = util.bytestring_path(f"{title}.mp3")
-                dest = os.path.join(album_dir, title_file)
-                if os.path.exists(syspath(dest)):
-                    track_no += 1
-                    continue
-                album_item_count -= 1
-                shutil.copy(syspath(src), syspath(dest))
-                mediafile = MediaFile(dest)
-                mediafile.update(
-                    {
-                        "artist": "artist",
-                        "albumartist": "album artist",
-                        "title": title,
-                        "album": album,
-                        "mb_albumid": None,
-                        "mb_trackid": None,
-                    }
-                )
-                mediafile.save()
-
-        config["import"]["quiet"] = True
-        config["import"]["autotag"] = False
-        config["import"]["resume"] = False
-
-        return ImportSessionFixture(
-            self.lib, loghandler=None, query=None, paths=[import_dir]
-        )
 
     # Library fixtures methods
 
@@ -296,20 +265,19 @@ class TestHelper:
 
         The item is attached to the database from `self.lib`.
         """
-        item_count = self._get_item_count()
         values_ = {
-            "title": "t\u00eftle {0}",
+            "title": "t\u00eftle {}",
             "artist": "the \u00e4rtist",
             "album": "the \u00e4lbum",
-            "track": item_count,
+            "track": 1,
             "format": "MP3",
         }
         values_.update(values)
-        values_["title"] = values_["title"].format(item_count)
+        values_["title"] = values_["title"].format(1)
         values_["db"] = self.lib
         item = Item(**values_)
         if "path" not in values:
-            item["path"] = "audio." + item["format"].lower()
+            item["path"] = f"audio.{item['format'].lower()}"
         # mtime needs to be set last since other assignments reset it.
         item.mtime = 12345
         return item
@@ -341,7 +309,7 @@ class TestHelper:
         item = self.create_item(**values)
         extension = item["format"].lower()
         item["path"] = os.path.join(
-            _common.RSRC, util.bytestring_path("min." + extension)
+            _common.RSRC, util.bytestring_path(f"min.{extension}")
         )
         item.add(self.lib)
         item.move(operation=MoveOperation.COPY)
@@ -356,7 +324,7 @@ class TestHelper:
         """Add a number of items with files to the database."""
         # TODO base this on `add_item()`
         items = []
-        path = os.path.join(_common.RSRC, util.bytestring_path("full." + ext))
+        path = os.path.join(_common.RSRC, util.bytestring_path(f"full.{ext}"))
         for i in range(count):
             item = Item.from_path(path)
             item.album = f"\u00e4lbum {i}"  # Check unicode paths
@@ -397,18 +365,14 @@ class TestHelper:
         return self.lib.add_album(items)
 
     def create_mediafile_fixture(self, ext="mp3", images=[]):
-        """Copies a fixture mediafile with the extension to a temporary
-        location and returns the path.
-
-        It keeps track of the created locations and will delete the with
-        `remove_mediafile_fixtures()`
+        """Copy a fixture mediafile with the extension to `temp_dir`.
 
         `images` is a subset of 'png', 'jpg', and 'tiff'. For each
         specified extension a cover art image is added to the media
         file.
         """
-        src = os.path.join(_common.RSRC, util.bytestring_path("full." + ext))
-        handle, path = mkstemp()
+        src = os.path.join(_common.RSRC, util.bytestring_path(f"full.{ext}"))
+        handle, path = mkstemp(dir=self.temp_dir)
         path = bytestring_path(path)
         os.close(handle)
         shutil.copyfile(syspath(src), syspath(path))
@@ -424,23 +388,8 @@ class TestHelper:
             mediafile.images = imgs
             mediafile.save()
 
-        if not hasattr(self, "_mediafile_fixtures"):
-            self._mediafile_fixtures = []
-        self._mediafile_fixtures.append(path)
-
         return path
 
-    def remove_mediafile_fixtures(self):
-        if hasattr(self, "_mediafile_fixtures"):
-            for path in self._mediafile_fixtures:
-                os.remove(syspath(path))
-
-    def _get_item_count(self):
-        if not hasattr(self, "__item_count"):
-            count = 0
-        self.__item_count = count + 1
-        return count
-
     # Running beets commands
 
     def run_command(self, *args, **kwargs):
@@ -453,7 +402,7 @@ class TestHelper:
         if hasattr(self, "lib"):
             lib = self.lib
         lib = kwargs.get("lib", lib)
-        beets.ui._raw_main(_convert_args(list(args)), lib)
+        beets.ui._raw_main(list(args), lib)
 
     def run_with_output(self, *args):
         with capture_stdout() as out:
@@ -462,16 +411,12 @@ class TestHelper:
 
     # Safe file operations
 
-    def create_temp_dir(self):
-        """Create a temporary directory and assign it into
-        `self.temp_dir`. Call `remove_temp_dir` later to delete it.
-        """
-        temp_dir = mkdtemp()
-        self.temp_dir = util.bytestring_path(temp_dir)
+    def create_temp_dir(self, **kwargs) -> str:
+        return mkdtemp(**kwargs)
 
     def remove_temp_dir(self):
         """Delete the temporary directory created by `create_temp_dir`."""
-        shutil.rmtree(syspath(self.temp_dir))
+        shutil.rmtree(self.temp_dir_path)
 
     def touch(self, path, dir=None, content=""):
         """Create a file at `path` with given content.
@@ -495,127 +440,226 @@ class TestHelper:
         return path
 
 
+# A test harness for all beets tests.
+# Provides temporary, isolated configuration.
+class BeetsTestCase(unittest.TestCase, TestHelper):
+    """A unittest.TestCase subclass that saves and restores beets'
+    global configuration. This allows tests to make temporary
+    modifications that will then be automatically removed when the test
+    completes. Also provides some additional assertion methods, a
+    temporary directory, and a DummyIO.
+    """
+
+    def setUp(self):
+        self.setup_beets()
+
+    def tearDown(self):
+        self.teardown_beets()
+
+
+class ItemInDBTestCase(BeetsTestCase):
+    """A test case that includes an in-memory library object (`lib`) and
+    an item added to the library (`i`).
+    """
+
+    def setUp(self):
+        super().setUp()
+        self.i = _common.item(self.lib)
+
+
+class PluginMixin(ConfigMixin):
+    plugin: ClassVar[str]
+    preload_plugin: ClassVar[bool] = True
+
+    def setup_beets(self):
+        super().setup_beets()
+        if self.preload_plugin:
+            self.load_plugins()
+
+    def teardown_beets(self):
+        super().teardown_beets()
+        self.unload_plugins()
+
+    def register_plugin(
+        self, plugin_class: type[beets.plugins.BeetsPlugin]
+    ) -> None:
+        beets.plugins._instances.append(plugin_class())
+
+    def load_plugins(self, *plugins: str) -> None:
+        """Load and initialize plugins by names.
+
+        Similar setting a list of plugins in the configuration. Make
+        sure you call ``unload_plugins()`` afterwards.
+        """
+        # FIXME this should eventually be handled by a plugin manager
+        plugins = (self.plugin,) if hasattr(self, "plugin") else plugins
+        self.config["plugins"] = plugins
+        beets.plugins.load_plugins()
+
+    def unload_plugins(self) -> None:
+        """Unload all plugins and remove them from the configuration."""
+        # FIXME this should eventually be handled by a plugin manager
+        beets.plugins.BeetsPlugin.listeners.clear()
+        beets.plugins.BeetsPlugin._raw_listeners.clear()
+        self.config["plugins"] = []
+        beets.plugins._instances.clear()
+
+    @contextmanager
+    def configure_plugin(self, config: Any):
+        self.config[self.plugin].set(config)
+        self.load_plugins(self.plugin)
+
+        yield
+
+        self.unload_plugins()
+
+
+class PluginTestCase(PluginMixin, BeetsTestCase):
+    pass
+
+
 class ImportHelper(TestHelper):
     """Provides tools to setup a library, a directory containing files that are
     to be imported and an import session. The class also provides stubs for the
     autotagging library and several assertions for the library.
     """
 
-    def setup_beets(self, disk=False):
-        super().setup_beets(disk)
+    default_import_config = {
+        "autotag": True,
+        "copy": True,
+        "hardlink": False,
+        "link": False,
+        "move": False,
+        "resume": False,
+        "singletons": False,
+        "timid": True,
+    }
+
+    lib: Library
+    importer: ImportSession
+
+    @cached_property
+    def import_path(self) -> Path:
+        import_path = self.temp_dir_path / "import"
+        import_path.mkdir(exist_ok=True)
+        return import_path
+
+    @cached_property
+    def import_dir(self) -> bytes:
+        return bytestring_path(self.import_path)
+
+    def setUp(self):
+        super().setUp()
+        self.import_media = []
         self.lib.path_formats = [
             ("default", os.path.join("$artist", "$album", "$title")),
             ("singleton:true", os.path.join("singletons", "$title")),
             ("comp:true", os.path.join("compilations", "$album", "$title")),
         ]
 
-    def _create_import_dir(self, count=3):
-        """Creates a directory with media files to import.
-        Sets ``self.import_dir`` to the path of the directory. Also sets
-        ``self.import_media`` to a list :class:`MediaFile` for all the files in
-        the directory.
+    def prepare_track_for_import(
+        self,
+        track_id: int,
+        album_path: Path,
+        album_id: int | None = None,
+    ) -> Path:
+        track_path = album_path / f"track_{track_id}.mp3"
+        shutil.copy(self.resource_path, track_path)
+        medium = MediaFile(track_path)
+        medium.update(
+            {
+                "album": f"Tag Album{f' {album_id}' if album_id else ''}",
+                "albumartist": None,
+                "mb_albumid": None,
+                "comp": None,
+                "artist": "Tag Artist",
+                "title": f"Tag Track {track_id}",
+                "track": track_id,
+                "mb_trackid": None,
+            }
+        )
+        medium.save()
+        self.import_media.append(medium)
+        return track_path
+
+    def prepare_album_for_import(
+        self,
+        item_count: int,
+        album_id: int | None = None,
+        album_path: Path | None = None,
+    ) -> list[Path]:
+        """Create an album directory with media files to import.
 
         The directory has following layout
-          the_album/
+          album/
             track_1.mp3
             track_2.mp3
             track_3.mp3
-
-        :param count:  Number of files to create
         """
-        self.import_dir = os.path.join(self.temp_dir, b"testsrcdir")
-        if os.path.isdir(syspath(self.import_dir)):
-            shutil.rmtree(syspath(self.import_dir))
+        if not album_path:
+            album_dir = f"album_{album_id}" if album_id else "album"
+            album_path = self.import_path / album_dir
 
-        album_path = os.path.join(self.import_dir, b"the_album")
-        os.makedirs(syspath(album_path))
+        album_path.mkdir(exist_ok=True)
 
-        resource_path = os.path.join(_common.RSRC, b"full.mp3")
+        return [
+            self.prepare_track_for_import(tid, album_path, album_id=album_id)
+            for tid in range(1, item_count + 1)
+        ]
 
-        metadata = {
-            "artist": "Tag Artist",
-            "album": "Tag Album",
-            "albumartist": None,
-            "mb_trackid": None,
-            "mb_albumid": None,
-            "comp": None,
-        }
-        self.media_files = []
-        for i in range(count):
-            # Copy files
-            medium_path = os.path.join(
-                album_path, bytestring_path("track_%d.mp3" % (i + 1))
-            )
-            shutil.copy(syspath(resource_path), syspath(medium_path))
-            medium = MediaFile(medium_path)
+    def prepare_albums_for_import(self, count: int = 1) -> None:
+        album_dirs = self.import_path.glob("album_*")
+        base_idx = int(str(max(album_dirs, default="0")).split("_")[-1]) + 1
 
-            # Set metadata
-            metadata["track"] = i + 1
-            metadata["title"] = "Tag Title %d" % (i + 1)
-            for attr in metadata:
-                setattr(medium, attr, metadata[attr])
-            medium.save()
-            self.media_files.append(medium)
-        self.import_media = self.media_files
+        for album_id in range(base_idx, count + base_idx):
+            self.prepare_album_for_import(1, album_id=album_id)
 
-    def _setup_import_session(
-        self,
-        import_dir=None,
-        delete=False,
-        threaded=False,
-        copy=True,
-        singletons=False,
-        move=False,
-        autotag=True,
-        link=False,
-        hardlink=False,
-    ):
-        config["import"]["copy"] = copy
-        config["import"]["delete"] = delete
-        config["import"]["timid"] = True
-        config["threaded"] = False
-        config["import"]["singletons"] = singletons
-        config["import"]["move"] = move
-        config["import"]["autotag"] = autotag
-        config["import"]["resume"] = False
-        config["import"]["link"] = link
-        config["import"]["hardlink"] = hardlink
-
-        self.importer = ImportSessionFixture(
+    def _get_import_session(self, import_dir: bytes) -> ImportSession:
+        return ImportSessionFixture(
             self.lib,
             loghandler=None,
             query=None,
-            paths=[import_dir or self.import_dir],
+            paths=[import_dir],
         )
 
-    def assert_file_in_lib(self, *segments):
-        """Join the ``segments`` and assert that this path exists in the
-        library directory.
-        """
-        self.assertExists(os.path.join(self.libdir, *segments))
+    def setup_importer(
+        self, import_dir: bytes | None = None, **kwargs
+    ) -> ImportSession:
+        self.config["import"].set_args({**self.default_import_config, **kwargs})
+        self.importer = self._get_import_session(import_dir or self.import_dir)
+        return self.importer
 
-    def assert_file_not_in_lib(self, *segments):
-        """Join the ``segments`` and assert that this path does not
-        exist in the library directory.
-        """
-        self.assertNotExists(os.path.join(self.libdir, *segments))
-
-    def assert_lib_dir_empty(self):
-        self.assertEqual(len(os.listdir(syspath(self.libdir))), 0)
+    def setup_singleton_importer(self, **kwargs) -> ImportSession:
+        return self.setup_importer(singletons=True, **kwargs)
 
 
-class ImportSessionFixture(importer.ImportSession):
+class AsIsImporterMixin:
+    def setUp(self):
+        super().setUp()
+        self.prepare_album_for_import(1)
+
+    def run_asis_importer(self, **kwargs):
+        importer = self.setup_importer(autotag=False, **kwargs)
+        importer.run()
+        return importer
+
+
+class ImportTestCase(ImportHelper, BeetsTestCase):
+    pass
+
+
+class ImportSessionFixture(ImportSession):
     """ImportSession that can be controlled programaticaly.
 
     >>> lib = Library(':memory:')
     >>> importer = ImportSessionFixture(lib, paths=['/path/to/import'])
-    >>> importer.add_choice(importer.action.SKIP)
-    >>> importer.add_choice(importer.action.ASIS)
-    >>> importer.default_choice = importer.action.APPLY
+    >>> importer.add_choice(importer.Action.SKIP)
+    >>> importer.add_choice(importer.Action.ASIS)
+    >>> importer.default_choice = importer.Action.APPLY
     >>> importer.run()
 
     This imports ``/path/to/import`` into `lib`. It skips the first
-    album and imports thesecond one with metadata from the tags. For the
+    album and imports the second one with metadata from the tags. For the
     remaining albums, the metadata from the autotagger will be applied.
     """
 
@@ -624,7 +668,7 @@ class ImportSessionFixture(importer.ImportSession):
         self._choices = []
         self._resolutions = []
 
-    default_choice = importer.action.APPLY
+    default_choice = importer.Action.APPLY
 
     def add_choice(self, choice):
         self._choices.append(choice)
@@ -638,7 +682,7 @@ class ImportSessionFixture(importer.ImportSession):
         except IndexError:
             choice = self.default_choice
 
-        if choice == importer.action.APPLY:
+        if choice == importer.Action.APPLY:
             return task.candidates[0]
         elif isinstance(choice, int):
             return task.candidates[choice - 1]
@@ -651,10 +695,6 @@ class ImportSessionFixture(importer.ImportSession):
 
     default_resolution = "REMOVE"
 
-    def add_resolution(self, resolution):
-        assert isinstance(resolution, self.Resolution)
-        self._resolutions.append(resolution)
-
     def resolve_duplicate(self, task, found_duplicates):
         try:
             res = self._resolutions.pop(0)
@@ -662,7 +702,7 @@ class ImportSessionFixture(importer.ImportSession):
             res = self.default_resolution
 
         if res == self.Resolution.SKIP:
-            task.set_choice(importer.action.SKIP)
+            task.set_choice(importer.Action.SKIP)
         elif res == self.Resolution.REMOVE:
             task.should_remove_duplicates = True
         elif res == self.Resolution.MERGE:
@@ -675,7 +715,7 @@ class TerminalImportSessionFixture(TerminalImportSession):
         super().__init__(*args, **kwargs)
         self._choices = []
 
-    default_choice = importer.action.APPLY
+    default_choice = importer.Action.APPLY
 
     def add_choice(self, choice):
         self._choices.append(choice)
@@ -697,134 +737,39 @@ class TerminalImportSessionFixture(TerminalImportSession):
         except IndexError:
             choice = self.default_choice
 
-        if choice == importer.action.APPLY:
+        if choice == importer.Action.APPLY:
             self.io.addinput("A")
-        elif choice == importer.action.ASIS:
+        elif choice == importer.Action.ASIS:
             self.io.addinput("U")
-        elif choice == importer.action.ALBUMS:
+        elif choice == importer.Action.ALBUMS:
             self.io.addinput("G")
-        elif choice == importer.action.TRACKS:
+        elif choice == importer.Action.TRACKS:
             self.io.addinput("T")
-        elif choice == importer.action.SKIP:
+        elif choice == importer.Action.SKIP:
             self.io.addinput("S")
-        elif isinstance(choice, int):
+        else:
             self.io.addinput("M")
             self.io.addinput(str(choice))
             self._add_choice_input()
-        else:
-            raise Exception("Unknown choice %s" % choice)
 
 
-class TerminalImportSessionSetup:
-    """Overwrites ImportHelper._setup_import_session to provide a terminal importer"""
+class TerminalImportMixin(IOMixin, ImportHelper):
+    """Provides_a terminal importer for the import session."""
 
-    def _setup_import_session(
-        self,
-        import_dir=None,
-        delete=False,
-        threaded=False,
-        copy=True,
-        singletons=False,
-        move=False,
-        autotag=True,
-    ):
-        config["import"]["copy"] = copy
-        config["import"]["delete"] = delete
-        config["import"]["timid"] = True
-        config["threaded"] = False
-        config["import"]["singletons"] = singletons
-        config["import"]["move"] = move
-        config["import"]["autotag"] = autotag
-        config["import"]["resume"] = False
+    io: _common.DummyIO
 
-        if not hasattr(self, "io"):
-            self.io = _common.DummyIO()
+    def _get_import_session(self, import_dir: bytes) -> importer.ImportSession:
         self.io.install()
-        self.importer = TerminalImportSessionFixture(
+        return TerminalImportSessionFixture(
             self.lib,
             loghandler=None,
             query=None,
             io=self.io,
-            paths=[import_dir or self.import_dir],
+            paths=[import_dir],
         )
 
 
-def generate_album_info(album_id, track_values):
-    """Return `AlbumInfo` populated with mock data.
-
-    Sets the album info's `album_id` field is set to the corresponding
-    argument. For each pair (`id`, `values`) in `track_values` the `TrackInfo`
-    from `generate_track_info` is added to the album info's `tracks` field.
-    Most other fields of the album and track info are set to "album
-    info" and "track info", respectively.
-    """
-    tracks = [generate_track_info(id, values) for id, values in track_values]
-    album = AlbumInfo(
-        album_id="album info",
-        album="album info",
-        artist="album info",
-        artist_id="album info",
-        tracks=tracks,
-    )
-    for field in ALBUM_INFO_FIELDS:
-        setattr(album, field, "album info")
-
-    return album
-
-
-ALBUM_INFO_FIELDS = [
-    "album",
-    "album_id",
-    "artist",
-    "artist_id",
-    "asin",
-    "albumtype",
-    "va",
-    "label",
-    "barcode",
-    "artist_sort",
-    "releasegroup_id",
-    "catalognum",
-    "language",
-    "country",
-    "albumstatus",
-    "media",
-    "albumdisambig",
-    "releasegroupdisambig",
-    "artist_credit",
-    "data_source",
-    "data_url",
-]
-
-
-def generate_track_info(track_id="track info", values={}):
-    """Return `TrackInfo` populated with mock data.
-
-    The `track_id` field is set to the corresponding argument. All other
-    string fields are set to "track info".
-    """
-    track = TrackInfo(
-        title="track info",
-        track_id=track_id,
-    )
-    for field in TRACK_INFO_FIELDS:
-        setattr(track, field, "track info")
-    for field, value in values.items():
-        setattr(track, field, value)
-    return track
-
-
-TRACK_INFO_FIELDS = [
-    "artist",
-    "artist_id",
-    "artist_sort",
-    "disctitle",
-    "artist_credit",
-    "data_source",
-    "data_url",
-]
-
-
+@dataclass
 class AutotagStub:
     """Stub out MusicBrainz album and track matcher and control what the
     autotagger returns.
@@ -835,47 +780,44 @@ class AutotagStub:
     GOOD = "GOOD"
     BAD = "BAD"
     MISSING = "MISSING"
-    """Generate an album match for all but one track
-    """
+    matching: str
 
     length = 2
-    matching = IDENT
 
     def install(self):
-        self.mb_match_album = autotag.mb.match_album
-        self.mb_match_track = autotag.mb.match_track
-        self.mb_album_for_id = autotag.mb.album_for_id
-        self.mb_track_for_id = autotag.mb.track_for_id
-
-        autotag.mb.match_album = self.match_album
-        autotag.mb.match_track = self.match_track
-        autotag.mb.album_for_id = self.album_for_id
-        autotag.mb.track_for_id = self.track_for_id
+        self.patchers = [
+            patch("beets.metadata_plugins.album_for_id", lambda *_: None),
+            patch("beets.metadata_plugins.track_for_id", lambda *_: None),
+            patch("beets.metadata_plugins.candidates", self.candidates),
+            patch(
+                "beets.metadata_plugins.item_candidates", self.item_candidates
+            ),
+        ]
+        for p in self.patchers:
+            p.start()
 
         return self
 
     def restore(self):
-        autotag.mb.match_album = self.mb_match_album
-        autotag.mb.match_track = self.mb_match_track
-        autotag.mb.album_for_id = self.mb_album_for_id
-        autotag.mb.track_for_id = self.mb_track_for_id
+        for p in self.patchers:
+            p.stop()
 
-    def match_album(self, albumartist, album, tracks, extra_tags):
+    def candidates(self, items, artist, album, va_likely):
         if self.matching == self.IDENT:
-            yield self._make_album_match(albumartist, album, tracks)
+            yield self._make_album_match(artist, album, len(items))
 
         elif self.matching == self.GOOD:
             for i in range(self.length):
-                yield self._make_album_match(albumartist, album, tracks, i)
+                yield self._make_album_match(artist, album, len(items), i)
 
         elif self.matching == self.BAD:
             for i in range(self.length):
-                yield self._make_album_match(albumartist, album, tracks, i + 1)
+                yield self._make_album_match(artist, album, len(items), i + 1)
 
         elif self.matching == self.MISSING:
-            yield self._make_album_match(albumartist, album, tracks, missing=1)
+            yield self._make_album_match(artist, album, len(items), missing=1)
 
-    def match_track(self, artist, title):
+    def item_candidates(self, item, artist, title):
         yield TrackInfo(
             title=title.replace("Tag", "Applied"),
             track_id="trackid",
@@ -885,31 +827,23 @@ class AutotagStub:
             index=0,
         )
 
-    def album_for_id(self, mbid):
-        return None
-
-    def track_for_id(self, mbid):
-        return None
-
     def _make_track_match(self, artist, album, number):
         return TrackInfo(
-            title="Applied Title %d" % number,
-            track_id="match %d" % number,
+            title=f"Applied Track {number}",
+            track_id=f"match {number}",
             artist=artist,
             length=1,
             index=0,
         )
 
     def _make_album_match(self, artist, album, tracks, distance=0, missing=0):
-        if distance:
-            id = " " + "M" * distance
-        else:
-            id = ""
+        id = f" {'M' * distance}" if distance else ""
+
         if artist is None:
             artist = "Various Artists"
         else:
-            artist = artist.replace("Tag", "Applied") + id
-        album = album.replace("Tag", "Applied") + id
+            artist = f"{artist.replace('Tag', 'Applied')}{id}"
+        album = f"{album.replace('Tag', 'Applied')}{id}"
 
         track_infos = []
         for i in range(tracks - missing):
@@ -920,8 +854,77 @@ class AutotagStub:
             album=album,
             tracks=track_infos,
             va=False,
-            album_id="albumid" + id,
-            artist_id="artistid" + id,
+            album_id=f"albumid{id}",
+            artist_id=f"artistid{id}",
             albumtype="soundtrack",
             data_source="match_source",
+            bandcamp_album_id="bc_url",
         )
+
+
+class AutotagImportTestCase(ImportTestCase):
+    matching = AutotagStub.IDENT
+
+    def setUp(self):
+        super().setUp()
+        self.matcher = AutotagStub(self.matching).install()
+        self.addCleanup(self.matcher.restore)
+
+
+class FetchImageHelper:
+    """Helper mixin for mocking requests when fetching images
+    with remote art sources.
+    """
+
+    @responses.activate
+    def run(self, *args, **kwargs):
+        super().run(*args, **kwargs)
+
+    IMAGEHEADER: dict[str, bytes] = {
+        "image/jpeg": b"\xff\xd8\xff\x00\x00\x00JFIF",
+        "image/png": b"\211PNG\r\n\032\n",
+        "image/gif": b"GIF89a",
+        # dummy type that is definitely not a valid image content type
+        "image/watercolour": b"watercolour",
+        "text/html": (
+            b"<!DOCTYPE html>\n<html>\n<head>\n</head>\n"
+            b"<body>\n</body>\n</html>"
+        ),
+    }
+
+    def mock_response(
+        self,
+        url: str,
+        content_type: str = "image/jpeg",
+        file_type: None | str = None,
+    ) -> None:
+        # Potentially return a file of a type that differs from the
+        # server-advertised content type to mimic misbehaving servers.
+        if file_type is None:
+            file_type = content_type
+
+        try:
+            # imghdr reads 32 bytes
+            header = self.IMAGEHEADER[file_type].ljust(32, b"\x00")
+        except KeyError:
+            # If we can't return a file that looks like real file of the requested
+            # type, better fail the test than returning something else, which might
+            # violate assumption made when writing a test.
+            raise AssertionError(f"Mocking {file_type} responses not supported")
+
+        responses.add(
+            responses.GET,
+            url,
+            content_type=content_type,
+            body=header,
+        )
+
+
+class CleanupModulesMixin:
+    modules: ClassVar[tuple[str, ...]]
+
+    @classmethod
+    def tearDownClass(cls) -> None:
+        """Remove files created by the plugin."""
+        for module in cls.modules:
+            clean_module_tempdir(module)
